#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.
MIT Licensed.
Contact at www.sinclair.bio
"""

# Built-in modules #

# First party modules #

# Third party modules #

################################################################################
# This provides extra information for certain packages #
extra_clues = {
    'apt': "It is likely the `apt` command is missing because we are not "
           "currently running on the Ubuntu or Debian operating system.",
    'gsed': "The gsed package can be installed with `brew install gnu-sed`"
            " on macOS."
}

# Make a detailed message #
msg_base = "The executable '%s' is required for this operation." \
           " Unfortunately it cannot be found anywhere in your $PATH." \
           " Either you need to install '%s'" \
           " or you need to fix the $PATH environment variable."

################################################################################
def check_cmd(cmd_name, exception=True, extra_msg=None):
    """
    Checks if the given executable is found in the path.
    Optionally, raises an exception if the executable `cmd_name` is not found.
    Optionally, print an extra message as a string alongside the exception.
    """
    # Use our own which implementation #
    from plumbing.common import which
    # Try to find it in the $PATH #
    if which(cmd_name, safe=True) is not None: return True
    # Prepare the message #
    msg = msg_base % (cmd_name, cmd_name)
    # Add any hints #
    if cmd_name in extra_clues: msg += '\n\n' + extra_clues.get(cmd_name)
    # Add a custom extra message #
    if extra_msg: msg += '\n\n' + extra_msg
    # Raise an exception #
    if exception: raise Exception(msg)
    # Return #
    return False


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.
MIT Licensed.
Contact at www.sinclair.bio
"""

# Built-in modules #
import os, sys

# Internal modules #
from autopaths.dir_path import DirectoryPath

# Third party modules #

###############################################################################
class GitRepo(DirectoryPath):
    """
    A git repository with some convenience methods.
    Requires at least git 2.7 (released January 5th, 2015)
    for all methods to work correctly.
    """

    def __bool__(self): return self.is_a_repos
    __nonzero__ = __bool__

    def __init__(self, path, empty=False):
        # Super #
        DirectoryPath.__init__(self, path)
        # The git directory #
        self.git_dir = self.path + '.git'
        # Check exists #
        if not empty and not self:
            raise Exception("No git repository at '%s'" % self.git_dir)
        # Default arguments #
        self.default = ["--git-dir=" + self.git_dir, "--work-tree=" + self.path]

    def git(self, *args, **kwargs):
        if os.name == "posix":
            import sh
            return sh.git(*args, **kwargs)
        if os.name == "nt":
            import pbs3
            return pbs3.Command("git")(*args, **kwargs)

    #------------------------------- Properties ------------------------------#
    @property
    def is_a_repos(self):
        """Check there is a `.git` directory and this is in fact a repo."""
        return os.path.exists(self.git_dir)

    @property
    def tag(self):
        """For instance: u'1.0.3-69-gf0c796d-dirty'"""
        tag = self.git(self.default + ["describe", "--tags", "--dirty",
                                       "--always"])
        return tag.strip('\n')

    @property
    def hash(self):
        """For instance: u'f0c796dae64a5a118d88e60523c011d535e8c476'"""
        sha1 = self.git(self.default + ["rev-parse", "HEAD"])
        return sha1.strip('\n')

    @property
    def short_hash(self):
        """For instance: u'f0c796d'"""
        sha1 = self.git(self.default + ["rev-parse", "--short", "HEAD"])
        return sha1.strip('\n')

    @property
    def branch(self):
        """For instance: u'master'"""
        result = self.git(self.default + ['symbolic-ref', '--short', 'HEAD'])
        return result.strip('\n')

    @property
    def branches(self):
        """All branches in a list"""
        result = self.git(self.default + ['branch', '-a', '--no-color'])
        return [l.strip(' *\n') for l in result.split('\n') if l.strip(' *\n')]

    @property
    def remote_branch(self):
        """For instance: u'origin/master'"""
        result = self.git(self.default + ['rev-parse', '--symbolic-full-name',
                                          '--abbrev-ref', '@{u}'])
        return result.strip('\n')

    @property
    def remote_url(self):
        """For instance: u'origin/master'"""
        result = self.git(self.default + ['remote', 'get-url', 'origin'])
        return result.strip('\n')

    #--------------------------------- Methods -------------------------------#
    def clone_from(self, remote_url, shell=False):
        """Clone it when it doesn't exist yet."""
        # Check empty #
        assert not self.is_a_repos
        # Command #
        command = ['clone', remote_url, self.path]
        # Show on shell #
        if shell: return self.git(command, _out=sys.stdout, _err=sys.stderr)
        else:     return self.git(command)

    def re_clone(self, repo_dir):
        """Clone again, somewhere else."""
        self.git('clone', self.remote_url, repo_dir)
        return GitRepo(repo_dir)

    def add(self, what):
        return self.git(self.default + ['add', what])

    def commit(self, message):
        return self.git(self.default + ['commit', '-m', message])

    def push(self, source=None, destination=None, tags=False, shell=False):
        # Command #
        command = self.default + ['push']
        # Tags #
        if tags: command.append('--tags')
        # Source and dest #
        if source:                 command.append(source)
        if source and destination: command.append(destination)
        # Show on shell #
        if shell: return self.git(command, _out=sys.stdout, _err=sys.stderr)
        else:     return self.git(command)

    def tag_head(self, tag):
        return self.git(self.default + ['tag', tag, 'HEAD'])

    def pull(self, shell=False, thread=False):
        # Command #
        command = self.default + ['pull']
        # Show on shell #
        if shell: return self.git(command, _out=sys.stdout,_err=sys.stderr,
                                           _bg=thread)
        else:     return self.git(command, _bg=thread)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.
MIT Licensed.
Contact at www.sinclair.bio

Development script to test some of the methods in `plumbing`
and try out different things. This script can safely be ignored
and is meant simply as a sandbox.

Typically you would run this file from a command line like this:

     ipython3 -i -- ~/deploy/plumbing/test/dev/tmp_code.py
"""

# Built-in modules #
import os

# Internal modules #
from plumbing.dependencies import check_setup_py

# Third party modules #

###############################################################################
path = os.path.expanduser("~/deploy/pacmill/setup.py")
check_setup_py(path)


#!/usr/bin/env python2
# -*- coding: utf-8 -*-

"""
Typically you would run this file from a command line like this:

     ipython.exe -i -- /deploy/plumbing/tests/database/access_db/test_access_db.py
"""

# Built-in module #
import inspect, os

# Internal modules #
from plumbing.databases.access_database import AccessDatabase
from autopaths.file_path import FilePath

# Third party modules #
import pandas

# Constants #
file_name = inspect.getframeinfo(inspect.currentframe()).filename
this_dir  = os.path.dirname(os.path.abspath(file_name)) + '/'

# Never modify the original #
orig_db    = FilePath(this_dir + 'orig.mdb')
testing_db = FilePath(this_dir + 'testing.mdb')
orig_db.copy(testing_db)

# The database #
db = AccessDatabase(testing_db)

# Test #
print(db.tables)
print(db['tblClassifierSets'])

# Create df #
df = pandas.DataFrame({'A': ['A0', 'A1', 'A2', 'A3'],
                       'B': ['B0', 'B1', 'B2', 'B3'],
                       'C': ['C0', 'C1', 'C2', 'C3'],
                       'D': ['D0', 'D1', 'D2', 'D3']},
                       index=[0, 1, 2, 3])

# Insert table #
db.insert_df('dataframe', df)

# Close #
db.close()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Script to test the functionality of the property_pickled decorator.

Written by Lucas Sinclair. MIT Licensed.

Call it like this:

    $ ipython3 -i ~/repos/plumbing/test/cache/property_pickled.py
"""

# Internal modules #
from plumbing.cache import property_pickled

###############################################################################
class Square:
    def __init__(self, size):
        self.size = size

    @property_pickled
    def area(self):
        print("Evaluating...")
        return self.size * self.size

###############################################################################
shape = Square(5)
print(shape.area)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.
MIT Licensed.
Contact at www.sinclair.bio
"""

# Built-in modules #
import os

# Internal modules #
from plumbing.cache import property_cached

# Third party modules #

###############################################################################
class InstanceEC2(object):
    """An EC2 instance provided by Amazon Web Services."""

    def __init__(self, instance_id, profile=None):
        # Basic attributes #
        self.instance_id  = instance_id
        self.profile      = profile
        # If we want a different profile set it in boto3 #
        import boto3
        if self.profile: boto3.setup_default_session(profile_name=self.profile)
        # Make the object we will use for queries #
        self.ec2 = boto3.client('ec2')

    #------------------------------ Properties -------------------------------#
    @property_cached
    def response(self):
        """Get the state of the instance"""
        return self.ec2.describe_instances(InstanceIds=[self.instance_id])

    @property
    def instance_info(self):
        """Get the information of the instance"""
        return self.response['Reservations'][0]['Instances'][0]

    @property
    def machine_off(self):
        """Is the machine stopped?"""
        return self.instance_info['State']['Code'] == 80

    @property
    def machine_on(self):
        """Is the machine running?"""
        return self.instance_info['State']['Code'] == 16

    @property
    def is_in_transition(self):
        """Is the machine transitioning?"""
        return not self.machine_off and not self.machine_on

    @property
    def dns(self):
        return self.instance_info['PublicDnsName']

    @property
    def tags(self):
        return self.instance_info['Tags']

    @property
    def instance_name(self):
        return [i['Value'] for i in self.tags if i['Key']=='Name'][0]

    #------------------------------- Methods ---------------------------------#
    def refresh_info(self):
        """Update the instance information that we cache."""
        del self.response

    def start_machine(self):
        """Start it up."""
        self.ec2.start_instances(InstanceIds=[self.instance_id])

    def rename(self, name):
        """Set the name of the machine."""
        self.ec2.create_tags(Resources = [self.instance_id],
                             Tags      = [{'Key':   'Name',
                                           'Value':  name}])
        self.refresh_info()

    def update_ssh_config(self, path="~/.ssh/config"):
        """Put the DNS into the ssh config file."""
        # Read the config file #
        import sshconf
        config = sshconf.read_ssh_config(os.path.expanduser(path))
        # In case it doesn't exist #
        if not config.host(self.instance_name): config.add(self.instance_name)
        # Add the new DNS #
        config.set(self.instance_name, Hostname=self.dns)
        # Write result #
        config.write(os.path.expanduser(path))


# Built-in modules #
import os, sqlite3

# Internal modules #
from itertools import islice

# First party modules #
from plumbing.color import Color

################################################################################
def convert_to_sql(dest, keys, values, sql_field_types=None):
    if sql_field_types is None: sql_field_types = {}
    with sqlite3.connect(dest) as connection:
        # Prepare #
        cursor = connection.cursor()
        fields = ','.join(['"' + f + '"' + ' ' + sql_field_types.get(f, 'text') for f in keys])
        cursor.execute("CREATE table 'data' (%s)" % fields)
        question_marks = '(' + ','.join(['?' for x in keys]) + ')'
        sql_command = "INSERT into 'data' values " + question_marks
        # Main loop #
        try:
            cursor.executemany(sql_command, values)
        except (ValueError, sqlite3.OperationalError, sqlite3.ProgrammingError, sqlite3.InterfaceError) as err:
            first_elem = islice(values, 0, 1)
            message1 = "The command <%s%s%s> on the database '%s' failed with error:\n %s%s%s"
            message1 = message1 % (Color.cyn, sql_command, Color.end, dest, Color.u_red, err, Color.end)
            message2 = "\n * %sThe bindings (%i) %s: %s \n * %sYou gave%s: %s"
            message2 = message2 % (Color.b_ylw, len(keys), Color.end, keys, Color.b_ylw, Color.end, values)
            message3 = "\n * %sFirst element (%i)%s: %s \n"
            message3 = message3 % (Color.b_ylw, len(first_elem) if first_elem else 0, Color.end, first_elem)
            raise Exception(message1 + message2 + message3)
        except KeyboardInterrupt as err:
            print("You interrupted the creation of the database. Committing everything done up to this point.")
            connection.commit()
            cursor.close()
            raise err
        # Index #
        try:
            cursor.execute("CREATE INDEX if not exists 'data_index' on 'data' (id)")
        except KeyboardInterrupt as err:
            print("You interrupted the creation of the index. Committing everything done up to this point.")
            connection.commit()
            cursor.close()
            raise err
        # Close #
        connection.commit()
        cursor.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.
MIT Licensed.
Contact at www.sinclair.bio
"""

# Built-in modules #
import logging

# Internal modules #

# Third party modules #

###############################################################################
def create_file_logger(name, path, file_level='debug', console_level='error'):
    # Create a custom logger #
    logger = logging.getLogger(name)
    # Console logger and file logger #
    c_handler = logging.StreamHandler()
    f_handler = logging.FileHandler(str(path), mode="w")
    # Choose the level of information for each #
    c_handler.setLevel(getattr(logging, console_level.upper()))
    f_handler.setLevel(getattr(logging, file_level.upper()))
    # Set the level of the logger itself to the lowest common denominator #
    #TODO: what if the console has a higher level than the file?
    logger.setLevel(getattr(logging, file_level.upper()))
    # Choose the format of each #
    c_format = '%(name)s - %(levelname)s - %(message)s'
    f_format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    c_format = logging.Formatter(c_format)
    f_format = logging.Formatter(f_format)
    c_handler.setFormatter(c_format)
    f_handler.setFormatter(f_format)
    # No need to display Exceptions on the console #
    class NoExceptionFilter(logging.Filter):
        def filter(self, record):
            return not record.getMessage() == 'Exception'
    c_handler.addFilter(NoExceptionFilter())
    # Add handlers to the logger #
    logger.addHandler(c_handler)
    logger.addHandler(f_handler)
    # Return #
    return logger



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.
MIT Licensed.
Contact at www.sinclair.bio
"""

# Built-in modules #
from datetime import datetime

# Internal modules #
from plumbing.color import Color

################################################################################
class Timer(object):
    """
    Useful for timing the different steps in a pipeline.

    Use it like this:

        from plumbing.timer import Timer
        timer = Timer()
        timer.print_start()
        timer.print_elapsed()
        timer.print_end()
        timer.print_total_elapsed()

    Or like this:

        with Timer():
            time.sleep(2)
    """

    def get_now(self):
        # What precision to use #
        if self.high_precision:
            return datetime.now()
        else:
            return datetime.now().replace(microsecond=0)

    def __init__(self, high_precision=False):
        self.high_precision = high_precision
        self.start_time = self.get_now()
        self.last_mark  = self.get_now()

    def print_start(self):
        # Parameters #
        time = self.start_time
        msg  = "Start at: %s" % time
        # Print #
        print(self.prefix + msg + self.suffix)

    def print_end(self):
        # Set #
        self.end_time = self.get_now()
        # Parameters #
        time = self.end_time
        msg  = "End at: %s" % time
        # Print #
        print(self.prefix + msg + self.suffix)

    def print_elapsed(self, reset=True):
        # Parameters #
        time = self.get_now() - self.last_mark
        msg  = "Elapsed time: %s" % time
        # Print #
        print(self.prefix + msg + self.suffix)
        # Set #
        if reset: self.last_mark = self.get_now()

    def print_total_elapsed(self, reset=True):
        # Parameters #
        time = self.get_now() - self.start_time
        msg = "Total elapsed time: %s"
        msg = msg % time
        # Print #
        print(self.prefix + msg + self.suffix)
        # Set #
        if reset: self.last_mark = self.get_now()

    @property
    def color(self):
        """
        Should we use color or not ?
        If we are not in a shell like ipython then not.
        """
        import __main__ as main
        return not hasattr(main, '__file__')

    @property
    def prefix(self): return "" if not self.color else Color.grn

    @property
    def suffix(self): return "" if not self.color else Color.end

    def __enter__(self):
        """Start the timer and print."""
        self.start_time = self.get_now()
        self.print_start()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Stop the timer and print."""
        self.print_end()
        self.print_total_elapsed()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.
MIT Licensed.
Contact at www.sinclair.bio
"""

# Built-in modules #
import sys

# Internal modules #
from plumbing.check_cmd_found import check_cmd

# Third party modules #
import sh

################################################################################
def get_apt_packages(all_packages, verbose=False):
    """
    Checks if the given aptitude packages are presently installed on the
    current system. Provide a list of strings as input.
    If one of the packages given as input is not installed, we will install it.
    """
    for pkg_name in all_packages:
        if check_apt_pkg(pkg_name): continue
        else: install_apt_pkg(pkg_name, verbose, verbose)

################################################################################
def check_apt_pkg(pkg_name, exception=False):
    """
    Checks if the given aptitude package is presently installed on the
    current system.
    Optionally, raises an exception if the package is not found.
    """
    # Call command #
    cmd = sh.Command('dpkg-query')
    result = cmd('-W', "--showformat='${Status}\\n'", pkg_name, _ok_code=[0,1])
    # Check result #
    if 'install ok installed' in result: return True
    # Make a message #
    msg = 'The package %s is not currently installed on this system. ' \
          'You can install it by typing: \n\n' \
          '    $ sudo apt install %s'
    msg = msg % (pkg_name, pkg_name)
    # Raise an exception #
    if exception: raise Exception(msg)
    # Return #
    return False

################################################################################
def install_apt_pkg(pkg_name, verbose=False, redirect=False):
    """
    Install a given apt package.
    Optionally, prints a status message before starting.
    Optionally, prints the output of apt on the current stdout.
    """
    # Optional message #
    if verbose: print("Installing apt package '%s'..." % pkg_name)
    # Check redirection #
    if redirect:
        sh.sudo.apt('install', '--yes', pkg_name, _out=sys.stdout, _err=sys.stderr)
    else:
        sh.sudo.apt('install', '--yes', pkg_name)

################################################################################
def check_apt_exists(exception=True):
    """
    Checks that the aptitude package manager is accessible on the current
    system.
    """
    msg = "You have encountered this error because we have attempted" \
          "the automatic installation of certain packages but are only" \
          "able do this using the `apt` package manager on supported OSes."
    return check_cmd('apt', exception, msg)



# Built-in modules #
import os, sqlite3

# Internal modules #
from plumbing.color      import Color
from plumbing.cache      import property_cached
from plumbing.scraping   import download_from_url
from autopaths.file_path import FilePath

# Third party modules #
import pandas

################################################################################
class SQLiteDatabase(FilePath):
    """A wrapper for an SQLite3 database."""

    def __init__(self, path,
                       factory   = None,
                       text_fact = None,
                       isolation = None,
                       retrieve  = None,
                       known_md5 = None):
        """
        * The path of the database comes first.
        * The factory option enables you to change how results are returned.
        * The text_fact option enables you to change text factory (useful for BLOB).
        * The Isolation source can be `None` for autocommit mode or one of:
          'DEFERRED', 'IMMEDIATE' or 'EXCLUSIVE'.
        * The retrieve option is an URL at which the database will be downloaded
          if it was not found at the `path` given.
        * The md5 option is used to check the integrity of a database."""
        self.path      = path
        self.text_fact = text_fact
        self.factory   = factory
        self.isolation = isolation
        self.retrieve  = retrieve
        self.known_md5 = known_md5
        self.prepared  = False

    def __repr__(self):
        """Called when evaluating ``print(database)``."""
        return '<%s object on "%s">' % (self.__class__.__name__, self.path)

    def __enter__(self):
        """Called when entering the 'with' statement."""
        return self

    def __exit__(self, errtype, value, traceback):
        """Called when exiting the 'with' statement.
        Enables us to close the database properly, even when
        exceptions are raised."""
        self.close()

    def __iter__(self):
        """Called when evaluating ``for x in database: pass``."""
        new_cursor = self.own_connection.cursor()
        new_cursor.execute('SELECT * from "%s";' % self.main_table)
        return new_cursor

    def __contains__(self, key):
        """Called when evaluating ``"P81239A" in database``."""
        command = 'SELECT EXISTS(SELECT 1 FROM "%s" WHERE id=="%s" LIMIT 1);'
        self.own_cursor.execute(command % (self.main_table, key))
        return bool(self.own_cursor.fetchone()[0])

    def __len__(self):
        """Called when evaluating ``len(database)``."""
        return self.count_entries()

    def __nonzero__(self):
        """Called when evaluating ``if database: pass``."""
        return True if len(self) != 0 else False

    def __getitem__(self, key):
        """Called when evaluating ``database[0] or database['P81239A']``."""
        return self.get_entry(key)

    # ------------------------------ Properties ----------------------------- #
    @property_cached
    def connection(self):
        """To be used externally by the user."""
        return self.new_connection()

    @property_cached
    def own_connection(self):
        """To be used internally in this object."""
        return self.new_connection()

    @property_cached
    def cursor(self):
        """To be used externally by the user."""
        return self.connection.cursor()

    @property_cached
    def own_cursor(self):
        """To be used internally in this object."""
        return self.own_connection.cursor()

    @property
    def tables(self):
        """The complete list of SQL tables."""
        self.own_connection.row_factory = sqlite3.Row
        self.own_cursor.execute('SELECT name from sqlite_master where type="table";')
        result = [x[0].encode('ascii') for x in self.own_cursor.fetchall()]
        self.own_connection.row_factory = self.factory
        return result

    @property_cached
    def main_table(self):
        return 'data'

    @property
    def columns(self):
        """The list of columns available in every entry."""
        return self.get_columns_of_table(self.main_table)

    @property
    def first(self):
        """Just the first entry."""
        return self.get_first()

    @property
    def last(self):
        """Just the last entry."""
        return self.get_last()

    @property
    def frame(self):
        """The main table as a blaze data structure. Not ready yet."""
        pass
        #return blaze.Data('sqlite:///%s::%s') % (self.path, self.main_table)

    # ------------------------------- Methods ------------------------------- #
    def new_connection(self):
        """Make a new connection."""
        if not self.prepared: self.prepare()
        con = sqlite3.connect(self.path, isolation_level=self.isolation)
        con.row_factory = self.factory
        if self.text_fact: con.text_factory = self.text_fact
        return con

    def prepare(self):
        """
        Check that the file exists, optionally downloads it.
        Checks that the file is indeed an SQLite3 database.
        Optionally check the MD5.
        """
        if not os.path.exists(self.path):
            if self.retrieve:
                print("Downloading SQLite3 database...")
                download_from_url(self.retrieve, self.path, progress=True)
            else: raise Exception("The file '" + self.path + "' does not exist.")
        self.check_format()
        if self.known_md5: assert self.known_md5 == self.md5
        self.prepared = True

    def check_format(self):
        if self.count_bytes == 0: return
        with open(self.path, 'r') as f: header = f.read(15)
        if header != 'SQLite format 3':
            raise Exception("The file '" + self.path + "' is not an SQLite database.")

    def create(self, columns=None, type_map=None, overwrite=False):
        """Create a new database with a certain schema."""
        # Check already exists #
        if self.count_bytes > 0:
            if overwrite: self.remove()
            else: raise Exception("File exists already at '%s'" % self)
        # If we want it empty #
        if columns is None:
            self.touch()
        # Make the table #
        else:
            self.add_table(self.main_table, columns=columns, type_map=type_map)

    def add_table(self, name, columns, type_map=None, if_not_exists=False):
        """
        Add add a new table to the database.  For instance you could do this:
        self.add_table('data', {'id':'integer', 'source':'text', 'pubmed':'integer'})
        """
        # Check types mapping #
        if type_map is None and isinstance(columns, dict): types = columns
        if type_map is None:                               types = {}
        # Safe or unsafe #
        if if_not_exists: query = 'CREATE TABLE IF NOT EXISTS "%s" (%s);'
        else:             query = 'CREATE table "%s" (%s);'
        # Do it #
        cols = ','.join(['"' + c + '"' + ' ' + types.get(c, 'text') for c in columns])
        self.own_cursor.execute(query % (self.main_table, cols))

    def execute(self, *args, **kwargs):
        """Convenience shortcut."""
        return self.cursor.execute(*args, **kwargs)

    def get_columns_of_table(self, table=None):
        """
        Return the list of columns for a particular table
        by querying the SQL for the complete list of column names.
        """
        # Check the table exists #
        if table is None: table = self.main_table
        if not table in self.tables: return []
        # A PRAGMA statement will implicitly issue a commit, don't use #
        self.own_cursor.execute('SELECT * from "%s" LIMIT 1;' % table)
        columns = [x[0] for x in self.own_cursor.description]
        self.cursor.fetchall()
        return columns

    def add(self, entries, table=None, columns=None, ignore=False):
        """
        Add entries to a table.
        The *entries* variable should be an iterable.
        """
        # Default table and columns #
        if table is None:   table   = self.main_table
        if columns is None: columns = self.get_columns_of_table(table)
        # Default columns #
        question_marks = ','.join('?' for c in columns)
        cols           = ','.join('"' + c + '"' for c in columns)
        ignore         = " OR IGNORE" if ignore else ""
        sql_command    = 'INSERT%s into "%s"(%s) VALUES (%s);'
        sql_command    = sql_command % (ignore, table, cols, question_marks)
        # Possible errors we want to catch #
        errors  = (sqlite3.OperationalError, sqlite3.ProgrammingError)
        errors += (sqlite3.IntegrityError, sqlite3.InterfaceError, ValueError)
        # Do it #
        try:
            new_cursor = self.own_connection.cursor()
            new_cursor.executemany(sql_command, entries)
        except errors as err:
            raise Exception(self.detailed_error(sql_command, columns, entries, err))
        except KeyboardInterrupt as err:
            print("You interrupted the data insertion.")
            print("Committing everything done up to this point.")
            self.own_connection.commit()
            raise err

    def detailed_error(self, sql_command, columns, entries, err):
        # The command #
        message1 = "\n\n The command \n <%s%s%s> \n on the database '%s' failed."
        params   = (Color.cyn, sql_command, Color.end, self)
        message1 = message1 % params
        # The bindings #
        message2 = "\n * %sThe bindings (%i) %s: %s \n * %sYou gave%s: %s"
        params   = (Color.b_ylw, len(columns), Color.end)
        params  += (columns, Color.b_ylw, Color.end, entries)
        message2 = message2 % params
        # The first element #
        message3 = "\n * %sExample element (%i)%s: %s \n"
        elem     = None
        if isinstance(entries, types.GeneratorType):
            try: elem = entries.next()
            except StopIteration: pass
        if isinstance(entries, (list, tuple)) and entries:
            elem = entries[0]
        params   =  (Color.b_ylw, len(elem) if elem else 0, Color.end, elem)
        message3 = message3 % params
        # The original error #
        message4 = "The original error was:\n %s%s%s \n" % (Color.u_red, err, Color.end)
        # Return #
        return message1 + message2 + message3 + message4

    def add_by_steps(self, entries_by_step, table=None, columns=None):
        """
        Add entries to the main table.
        The *entries* variable should be an iterable yielding iterables.
        """
        for entries in entries_by_step: self.add(entries, table=table, columns=columns)

    def count_entries(self, table=None):
        """How many rows in a table."""
        if table is None: table = self.main_table
        self.own_cursor.execute('SELECT COUNT(1) FROM "%s";' % table)
        return int(self.own_cursor.fetchone()[0])

    def index(self, column='id', table=None):
        if table is None: table = self.main_table
        index_name = table + '_index'
        try:
            command = 'CREATE INDEX if not exists "%s" on "%s" (%s);'
            command = command % (index_name, table, column)
            self.own_cursor.execute(command)
        except KeyboardInterrupt as err:
            print("You interrupted the creation of the index. Not committing.")
            raise err

    def get_first(self, table=None):
        """Just the first entry."""
        if table is None: table = self.main_table
        query = 'SELECT * FROM "%s" LIMIT 1;' % table
        return self.own_cursor.execute(query).fetchone()

    def get_last(self, table=None):
        """Just the last entry."""
        if table is None: table = self.main_table
        query = 'SELECT * FROM "%s" ORDER BY ROWID DESC LIMIT 1;' % table
        return self.own_cursor.execute(query).fetchone()

    def get_number(self, num, table=None):
        """Get a specific entry by its number."""
        if table is None: table = self.main_table
        self.own_cursor.execute('SELECT * from "%s" LIMIT 1 OFFSET %i;' % (self.main_table, num))
        return self.own_cursor.fetchone()

    def get(self, table, column, key): return self.get_entry(key, column, table)
    def get_entry(self, key, column=None, table=None):
        """Get a specific entry."""
        if table is None:  table  = self.main_table
        if column is None: column = "id"
        if isinstance(key, basestring): key = key.replace("'","''")
        query = 'SELECT * from "%s" where "%s"=="%s" LIMIT 1;'
        query = query % (table, column, key)
        self.own_cursor.execute(query)
        return self.own_cursor.fetchone()

    def vacuum(self):
        """Compact the database, remove old transactions."""
        self.own_cursor.execute("VACUUM")

    def close(self):
        self.cursor.close()
        self.connection.close()
        self.own_cursor.close()
        self.own_connection.close()

    def insert_df(self, table_name, df):
        """Create a table and populate it with data from a dataframe."""
        df.to_sql(table_name, con=self.own_connection)

    # ----------------------------- Unfinished ------------------------------ #
    def add_column(self, name, kind=None, table=None):
        """Add add a new column to a table."""
        pass

    def open(self):
        """Reopen a database that was closed, useful for debugging."""
        pass

    def uniquify(self, name, kind=None, table=None):
        """Remove entries that have duplicate values on a specific column."""
        query = 'DELETE from "data" WHERE rowid not in (select min(rowid) from data group by id);'
        pass

    def get_and_order(self, ids, column=None, table=None):
        """Get specific entries and order them in the same way."""
        command = """
        SELECT rowid, * from "data"
        WHERE rowid in (%s)
        ORDER BY CASE rowid
        %s
        END;
        """
        ordered = ','.join(map(str,ids))
        rowids  = '\n'.join("WHEN '%s' THEN %s" % (row,i) for i,row in enumerate(ids))
        command = command % (ordered, rowids)
        # This could have worked but sqlite3 was too old on the server
        # ORDER BY instr(',%s,', ',' || id || ',')

    # ---------------------------- Multidatabase ---------------------------- #
    def import_table(self, source, table_name):
        """Copy a table from another SQLite database to this one."""
        query = "SELECT * FROM `%s`" % table_name.lower()
        df = pandas.read_sql(query, source.connection)
        df.to_sql(table_name, con=self.own_connection)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Script to test the functionality of the scraping module.

Written by Lucas Sinclair. MIT Licensed.

Call it like this:

    $ ipython3 -i ~/repos/plumbing/test/scraping/test_user_agent.py
"""

# Internal modules #

###############################################################################
url = "http://example.com/index.html"

#from plumbing.scraping import retrieve_from_url
#print(retrieve_from_url(url, user_agent=1))

from plumbing.scraping import download_from_url
print(download_from_url(url, '~/test/example.com/extra_level/'))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Script to test the functionality of the scraping module.

Written by Lucas Sinclair. MIT Licensed.

Call it like this:

    $ ipython3 -i ~/repos/plumbing/test/scraping/test_user_agent.py
"""

# Internal modules #
from plumbing.scraping.browser import download_via_browser

# Constants #
url = "http://speedtest.tele2.net/1MB.zip"
destination = '~/test2/'

###############################################################################
download_via_browser(url, destination)

url = "http://example.com/"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Built-in modules #

# Internal modules #

# First party modules #
from plumbing.graphs import Graph

# Third party modules #

###############################################################################
class SoloLegend(Graph):
    """
    A figure that contains no plot, only a legend, for composition purposes
    with the other graphs.
    """

    short_name = "legend"
    n_col      = 1
    capitalize = True

    @property
    def label_to_color(self):
        raise NotImplementedError()

    def plot(self, **kwargs):
        # Import #
        import matplotlib
        from matplotlib import pyplot
        # Plot #
        fig  = pyplot.figure()
        axes = fig.add_subplot(111)
        # Get each item #
        items = self.label_to_color.items()
        # Modify the labels #
        if self.capitalize: items = [(k.capitalize(), v) for k,v in items]
        # Create the patches #
        patches = [matplotlib.patches.Patch(color=v, label=k) for k,v in items]
        # Suppress a warning #
        import warnings
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            leg = fig.legend(handles   = patches,
                             borderpad = 1,
                             prop      = {'size': 20},
                             frameon   = True,
                             shadow    = True,
                             loc       = 'center',
                             ncol      = self.n_col,
                             fancybox  = True)
        # Remove the axes #
        axes.axis('off')
        # Find the bounding box to remove useless white space #
        import numpy
        fig.canvas.draw()
        expand = [-10, -10, 10, 10]
        bbox   = leg.get_window_extent()
        bbox   = bbox.from_extents(*(bbox.extents + numpy.array(expand)))
        bbox   = bbox.transformed(fig.dpi_scale_trans.inverted())
        # Save #
        self.dpi  = 'figure'
        self.bbox = bbox
        self.save_plot(**kwargs)
        # Return for display in notebooks for instance #
        return fig


# Built-in modules #
import os, time, inspect, getpass
from collections import OrderedDict

# Internal modules #
from plumbing.common import split_thousands, camel_to_snake
from plumbing.cache  import property_cached

# First party modules #
from autopaths           import Path
from autopaths.file_path import FilePath

################################################################################
class Graph(object):
    """
    A nice class to make graphs with matplotlib. Example usage:

        from plumbing.graphs import Graph

        class RegressionGraph(Graph):
            formats = ('pdf', 'svg')
            def plot(self, **kwargs):
                fig = pyplot.figure()
                seaborn.regplot(self.x_data, self.y_data, fit_reg=True);
                self.save_plot(fig, **kwargs)

        for x_name in x_names:
            graph            = PearsonGraph(short_name = x_name)
            graph.title      = "Regression between y and '%s'" % (x_name)
            graph.x_data     = x_data[x_name]
            graph.y_data     = y_data
            graph.plot()
    """

    default_params = OrderedDict((
        ('width'  , None),
        ('height' , None),
        ('bottom' , None),
        ('top'    , None),
        ('left'   , None),
        ('right'  , None),
        ('x_grid' , None), # Vertical lines
        ('y_grid' , None), # Horizontal lines
        ('x_scale', None),
        ('y_scale', None),
        ('x_label', None),
        ('y_label', None),
        ('x_labels_rot', None),
        ('x_labels_size', None),
        ('y_labels_size', None),
        ('title'  , None),
        ('y_lim_min', None), # Minimum (ymax - ymin) after autoscale
        ('x_lim_min', None), # Minimum (xmax - xmin) after autoscale
        ('sep'    , ()),
        ('formats', ('pdf',)),
        ('close'  , True),
        ('dpi'    , None),
        ('bbox'   , None),
        ('remove_frame', None),
    ))

    def __repr__(self):
        return '<%s graph "%s">' % (self.__class__.__name__, self.short_name)

    def __bool__(self): return bool(self.path)
    __nonzero__ = __bool__

    def __init__(self, parent=None, base_dir=None, short_name=None):
        # Save parent #
        self.parent = parent
        # If we got a file as base_dir #
        if isinstance(base_dir, FilePath):
            self.base_dir = base_dir.directory
            short_name    = base_dir.short_prefix
        # If no parent and no directory given get the calling script #
        if base_dir is None and parent is None:
            file_name     = os.path.abspath((inspect.stack()[1])[1])
            self.base_dir = os.path.dirname(os.path.abspath(file_name)) + '/'
            self.base_dir = Path(self.base_dir)
        # If no directory given but a parent is present we can guess #
        if base_dir is None:
            if hasattr(self.parent, 'p'):
                self.base_dir = self.parent.p.graphs_dir
            elif hasattr(self.parent, 'paths'):
                self.base_dir = self.parent.paths.graphs_dir
            elif hasattr(self.parent, 'autopaths'):
                self.base_dir = self.parent.autopaths.graphs_dir
            elif hasattr(self.parent, 'graphs_dir'):
                self.base_dir = self.parent.graphs_dir
            else:
                raise Exception("Please specify a `base_dir` for this graph.")
        else:
            self.base_dir = Path(base_dir)
        # Make sure the directory exists #
        self.base_dir.create_if_not_exists()
        # Short name #
        if short_name: self.short_name = short_name
        # Use the parents name or the base class name #
        if not hasattr(self, 'short_name'):
            if hasattr(self.parent, 'short_name'):
                self.short_name = self.parent.short_name
            else:
                self.short_name = camel_to_snake(self.__class__.__name__)

    @property_cached
    def path(self):
        return Path(self.base_dir + self.short_name + '.pdf')

    def __call__(self, *args, **kwargs):
        """
        Plot the graph if it doesn't exist. Then return the path to it.
        Force the re-runing with rerun=True.
        """
        if not self or kwargs.get('rerun'): self.plot(*args, **kwargs)
        return self.path

    def save_plot(self, fig=None, axes=None, **kwargs):
        # Import #
        from matplotlib import pyplot
        # Missing figure #
        if fig is None:   fig = pyplot.gcf()
        # Missing axes #
        if axes is None: axes = pyplot.gca()
        # Parameters #
        self.params = {}
        for key in self.default_params:
            if key in kwargs:                          self.params[key] = kwargs[key]
            elif hasattr(self, key):                   self.params[key] = getattr(self, key)
            elif self.default_params[key] is not None: self.params[key] = self.default_params[key]
        # Backwards compatibility #
        if kwargs.get('x_log', False): self.params['x_scale'] = 'symlog'
        if kwargs.get('y_log', False): self.params['y_scale'] = 'symlog'
        # Log #
        if 'x_scale' in self.params: axes.set_xscale(self.params['x_scale'])
        if 'y_scale' in self.params: axes.set_yscale(self.params['y_scale'])
        # Axis limits #
        if 'x_min' in self.params: axes.set_xlim(self.params['x_min'], axes.get_xlim()[1])
        if 'x_max' in self.params: axes.set_xlim(axes.get_xlim()[0], self.params['x_max'])
        if 'y_min' in self.params: axes.set_ylim(self.params['y_min'], axes.get_ylim()[1])
        if 'y_max' in self.params: axes.set_ylim(axes.get_ylim()[0], self.params['y_max'])
        # Minimum delta on axis limits #
        if 'y_lim_min' in self.params:
            top, bottom = axes.get_ylim()
            minimum     = self.params['y_lim_min']
            delta       = top - bottom
            if delta < minimum:
                center = bottom + delta/2
                axes.set_ylim(center - minimum/2, center + minimum/2)
        # Title #
        title = self.params.get('title', False)
        if title: axes.set_title(title)
        # Axes labels  #
        if self.params.get('x_label'): axes.set_xlabel(self.params['x_label'])
        if self.params.get('y_label'): axes.set_ylabel(self.params['y_label'])
        # Set height and width #
        if self.params.get('width'):  fig.set_figwidth(self.params['width'])
        if self.params.get('height'): fig.set_figheight(self.params['height'])
        # Adjust #
        if self.params.get('bottom'):
            fig.subplots_adjust(hspace=0.0, bottom = self.params['bottom'], top   = self.params['top'],
                                            left   = self.params['left'],   right = self.params['right'])
        # Grid #
        if 'x_grid' in self.params:
            if self.params['x_grid']: axes.xaxis.grid(True, linestyle=':')
            else: axes.xaxis.grid(False)
        if 'y_grid' in self.params:
            if self.params['y_grid']: axes.yaxis.grid(True, linestyle=':')
            else: axes.yaxis.grid(False)
        # Frame #
        if 'remove_frame' in self.params:
            axes.spines["top"].set_visible(False)
            axes.spines["right"].set_visible(False)
        # Data and source extra text #
        if hasattr(self, 'dev_mode') and self.dev_mode is True:
            fig.text(0.99, 0.98, time.asctime(), horizontalalignment='right')
            job_name = os.environ.get('SLURM_JOB_NAME', 'Unnamed')
            user_msg = 'user: %s, job: %s' % (getpass.getuser(), job_name)
            fig.text(0.01, 0.98, user_msg, horizontalalignment='left')
        # Nice digit grouping #
        import matplotlib
        if 'x' in self.params['sep']:
            separate = lambda x,pos: split_thousands(x)
            axes.xaxis.set_major_formatter(matplotlib.ticker.FuncFormatter(separate))
        if 'y' in self.params['sep']:
            separate = lambda y,pos: split_thousands(y)
            axes.yaxis.set_major_formatter(matplotlib.ticker.FuncFormatter(separate))
        # Add custom labels #
        if 'x_labels' in self.params: axes.set_xticklabels(self.params['x_labels'])
        if 'x_labels_rot' in self.params:
            pyplot.setp(axes.xaxis.get_majorticklabels(), rotation=self.params['x_labels_rot'])
        # Adjust font size #
        if 'x_labels_size' in self.params:
            pyplot.setp(axes.xaxis.get_majorticklabels(), fontsize=self.params['x_labels_size'])
        if 'y_labels_size' in self.params:
            pyplot.setp(axes.yaxis.get_majorticklabels(), fontsize=self.params['y_labels_size'])
        # Possibility to overwrite path #
        if 'path' in self.params:   path = FilePath(self.params['path'])
        elif hasattr(self, 'path'): path = FilePath(self.path)
        else:                       path = FilePath(self.short_name + '.pdf')
        # The arguments to save #
        save_args = {}
        if 'dpi'  in self.params: save_args['dpi']  = self.params['dpi']
        if 'bbox' in self.params: save_args['bbox_inches'] = self.params['bbox']
        # Save it as different formats #
        for ext in self.params['formats']:
            fig.savefig(path.replace_extension(ext), **save_args)
        # Close it #
        if self.params['close']: pyplot.close(fig)

    def plot_and_save(self, **kwargs):
        """
        Used when the plot method defined does not create a figure nor calls save_plot
        Then the plot method has to use self.fig.
        """
        from matplotlib import pyplot
        self.fig = pyplot.figure()
        self.plot()
        self.axes = pyplot.gca()
        self.save_plot(self.fig, self.axes, **kwargs)
        pyplot.close(self.fig)

    def plot(self, bins=250, **kwargs):
        """An example plot function. You have to subclass this method."""
        # Import #
        import numpy
        # Data #
        counts = [sum(map(len, b.contigs)) for b in self.parent.bins]
        # Linear bins in logarithmic space #
        if 'log' in kwargs.get('x_scale', ''):
            start, stop = numpy.log10(1), numpy.log10(max(counts))
            bins = list(numpy.logspace(start=start, stop=stop, num=bins))
            bins.insert(0, 0)
        # Plot #
        from matplotlib import pyplot
        fig = pyplot.figure()
        pyplot.hist(counts, bins=bins, color='gray')
        axes = pyplot.gca()
        # Information #
        title = 'Distribution of the total nucleotide count in the bins'
        axes.set_title(title)
        axes.set_xlabel('Number of nucleotides in a bin')
        axes.set_ylabel('Number of bins with that many nucleotides in them')
        # Save it #
        self.save_plot(fig, axes, **kwargs)
        pyplot.close(fig)
        # For convenience #
        return self

    def save_anim(self, fig, animate, init, bitrate=10000, fps=30):
        """Not functional -- TODO."""
        from matplotlib import animation
        anim = animation.FuncAnimation(fig, animate, init_func=init, frames=360, interval=20)
        FFMpegWriter = animation.writers['ffmpeg']
        writer = FFMpegWriter(bitrate= bitrate, fps=fps)
        # Save #
        self.avi_path = self.base_dir + self.short_name + '.avi'
        anim.save(self.avi_path, writer=writer, codec='x264')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.
MIT Licensed.
Contact at www.sinclair.bio
"""

# Imports #
from setuptools import setup, find_namespace_packages
from os import path

# Load the contents of the README file #
this_dir = path.abspath(path.dirname(__file__))
readme_path = path.join(this_dir, 'README.md')
with open(readme_path, encoding='utf-8') as handle: readme = handle.read()

# Call setup #
setup(
    name             = 'plumbing',
    version          = '2.10.4',
    description      = 'Helps with plumbing-type programing in python.',
    license          = 'MIT',
    url              = 'http://github.com/xapple/plumbing/',
    author           = 'Lucas Sinclair',
    author_email     = 'lucas.sinclair@me.com',
    packages         = find_namespace_packages(exclude=['testing']),
    install_requires = ['autopaths>=1.4.12', 'pandas', 'retry', 'requests'],
    extras_require   = {'pretty_now':   ['tzlocal'],
                        'graphs':       ['matplotlib', 'numpy'],
                        'check_module': ['packaging']},
    python_requires  = ">=3.8",
    long_description = readme,
    long_description_content_type = 'text/markdown',
    include_package_data = True,
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.
MIT Licensed.
Contact at www.sinclair.bio
"""

# Built-in modules #
import os, shutil, time
import urllib.parse

# Internal modules #
from plumbing.scraping import handle_destination

# First party modules #
from autopaths import Path

###############################################################################
def make_driver():
    """
    Create a headless webdriver with selenium.

    Note: #TODO Change the download dir to a temp dir #
    Otherwise it will append (1) and (2) etc. to the filename before the
    extension e.g. 'data.zip (1).crdownload'
    """
    # Import #
    from selenium import webdriver
    # Paths #
    chrome_driver_path = shutil.which('chromedriver')
    # Start a service #
    service = webdriver.chrome.service.Service(chrome_driver_path)
    service.start()
    # Add options #
    options = webdriver.ChromeOptions()
    options.add_argument('--headless')
    options = options.to_capabilities()
    # Create the driver #
    driver = webdriver.Remote(service.service_url, options)
    # #TODO Change the download dir to a temp dir #
    dl_dir = Path(os.getcwd() + '/')
    # Return #
    return driver, dl_dir

###############################################################################
def download_via_browser(url,
                         destination = None,
                         uncompress  = True,
                         timeout     = 120):
    """
    Save the resource as a file on disk by running a full browser to avoid
    being blocked by the server.
    Note: While the file is downloading, it will have ".crdownload" appended
    to it.
    """
    # Choose the right option for destination #
    destination = handle_destination(url, destination)
    # Make a driver if we don't already have one #
    driver, download_dir = make_driver()
    # Get resource #
    driver.get(url)
    # This should remove the query part of the URL #
    predicted_name = url.split("/")[-1].split("?")[0]
    # But we need to convert things like %20 to a space #
    predicted_name = urllib.parse.unquote(predicted_name)
    # It should land in the downloads directory #
    result = download_dir + predicted_name
    # We have no way to know when it finishes #
    start_time = time.time()
    while True:
        time.sleep(0.5)
        if result.exists: break
        if time.time() - start_time > timeout:
            msg = "The file should have been at '%s'." % result
            raise Exception("The timeout was exceeded. " + msg)
    # Let's move it #
    result.move_to(destination, overwrite=True)
    # Uncompress #
    if uncompress:
        with open(destination, 'rb') as f: header = f.read(4)
        if header == b"PK\x03\x04": destination.unzip_to(inplace=True)
    # Return #
    return destination
